@@ -38,6 +38,9 @@ netplan_parser_load_yaml(NetplanParser* npp, const char* filename, NetplanError*
 NETPLAN_PUBLIC gboolean
 netplan_parser_load_yaml_from_fd(NetplanParser* npp, int input_fd, NetplanError** error);
 
+NETPLAN_PUBLIC gboolean
+netplan_parser_load_yaml_hierarchy(NetplanParser* npp, const char* rootdir, NetplanError** error);
+
 NETPLAN_PUBLIC gboolean
 netplan_parser_load_nullable_fields(NetplanParser* npp, int input_fd, NetplanError** error);
 
@@ -51,7 +54,7 @@ netplan_state_import_parser_results(NetplanState* np_state, NetplanParser* npp,
  * <constraint> only. */
 NETPLAN_PUBLIC gboolean
 netplan_parser_load_nullable_overrides(
-    NetplanParser* npp, int input_fd, const char* constraint, GError** error);
+    NetplanParser* npp, int input_fd, const char* constraint, NetplanError** error);
 
 /********** Old API below this ***********/
 
